/*
Copyright (c) 1998 David Kuehling

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice (including the next
paragraph) shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

/***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
// #include <dir.h>
/***************************************************************************/

#define MAXRECURSION 500
#define VERSION_STR  "1.03"

/***************************************************************************/
typedef struct ARGLIST
{
  char           *Name;
  int            Ellipse;
  struct ARGLIST *Next;
} ARGLIST;

typedef struct MACRO
{
  char          *Name,
                *Content;
  int           LocalRecursion;
  ARGLIST       *ArgList;
  struct  MACRO *Next;
} MACRO;

typedef struct INPUTSTACK
{
  char              *File;
  int               LineNum,
                    LocalMacros,
                    LabelRead;

  FILE              *Hdl;
  char              *Str;

  struct INPUTSTACK *Last;
} INPUTSTACK;

typedef struct LOCALSTACK
{
  int               Prepend,
                    LocalNum;
  struct LOCALSTACK *Last;
} LOCALSTACK;

typedef struct LOCAL
{
  char              *OrigName,
                    *LocalName;
  struct LOCAL      *Next;
} LOCAL;
/***************************************************************************/


/***************************************************************************/
char       *IncludePath = NULL,     /* path for include files #include <...> */
           *SourcePath = NULL,      /* path for include files #include "..." */
           *OutFile    = NULL,      /* name of the output file */
           Token[200000];           /* current token */


int        Recursion = 0,           /* current recursion depth */
           InputEnd = 0,            /* the complete input done? */
           IfLevel = 0,             /* number of nested #ifdef/#ifndef s*/
           LocalPrependNum = 0;     /* number that is prepended to local symbols */

MACRO      *MacroList = NULL;       /* pointer to linked macro list */
LOCAL      *LocalList = NULL;       /* pointer to linked local symbols list */
LOCALSTACK *Locals    = NULL;       /* stack for local symbols */
INPUTSTACK *Input     = NULL;       /* stack for input (macro/file etc.) */
FILE       *OutHdl;                 /* handle of output file */
/***************************************************************************/


/**
 * Allocates memory for the string, and prints an error message on error
 */
void *Xstrdup (char *str)
{
  void *ptr;
  if ((ptr = strdup (str)) == NULL)
  {
    fputs ("\nError: not enough DPMI memory.\n\n", stderr);
    exit (1);
  }
  return (ptr);
}

/**
 * Allocates memory and prints an error message on error
 */
void *Xmalloc (long bytes)
{
  void *ptr;
  if ((ptr = malloc (bytes)) == NULL)
  {
    fputs ("\nError: not enough DPMI memory.\n\n", stderr);
    exit (1);
  }
  return (ptr);
}

/**
 * Prints the error, and exits with exitcode 1
 */
void Error (char *format, ...)
{
  va_list arg;

  va_start(arg, format);
  fprintf(stderr, "\n%s(%i) Error: ", Input->File, Input->LineNum);
  vfprintf(stderr, format, arg);
  va_end(arg);
  fputs ("\n\n", stderr);
  fclose (OutHdl);
  remove (OutFile);
  exit (1);
}

/**
 * Prints a warning
 */
void Warning (char *format, ...)
{
  va_list arg;

  va_start(arg, format);
  fprintf(stderr, "%s(%i) Warning: ", Input->File, Input->LineNum);
  vfprintf(stderr, format, arg);
  va_end(arg);
  fputs ("\n", stderr);
}

/***************************************************************************
 * Routines for processing macros
 */

/**
 * Returns the address of a macro
 */
MACRO *getMacro (char *name)
{
  MACRO *I;

  for (I = MacroList; I != NULL; I = I->Next)
    if (!strcmp (name, I->Name) &&
        (I->LocalRecursion == Recursion || !I->LocalRecursion))
      return (I);

  return (NULL);
}

/**
 * Frees the contents of a macro
 */
void FreeMacroContents (MACRO *macro)
{
  ARGLIST *Iarg,
          *kill;

  free (macro->Name);
  free (macro->Content);

  for (Iarg = macro->ArgList; Iarg != NULL;)
  {
    free (Iarg->Name);
    kill = Iarg;
    Iarg = Iarg->Next;
    free (kill);
  }
}

/**
 * Frees the memory, used by a macro
 */
void FreeMacro (MACRO *macro)
{
  FreeMacroContents (macro);
  free (macro);
}

/**
 * Adds a macro to the macro list
 */
void NewMacro (char *name, char *content, ARGLIST *arglist)
{
  MACRO *macro;

  if (macro = getMacro (name))
  {
    if (macro->LocalRecursion)
      Error ("redefinition of macro argument `%s', name");
    else
    {
      Warning ("`%s' redefined", name);
      FreeMacroContents (macro);
    }
  }
  else
  {
    macro = MacroList;
    MacroList = Xmalloc (sizeof (MACRO));
    MacroList->Next = macro;
    macro = MacroList;
  }
  macro->Name = Xstrdup (name);
  macro->Content = Xstrdup (content);
  macro->ArgList = arglist;
  macro->LocalRecursion = 0;
}

/**
 * Adds an equate-macro to the macro list
 */
void NewEquate (char *name, char *content)
{
  MACRO *macro;

  if (macro = getMacro (name))
  {
    if (macro->LocalRecursion)
      Error ("redefinition of macro argument `%s', name");
    FreeMacroContents (macro);
  }
  else
  {
    macro = MacroList;
    MacroList = Xmalloc (sizeof (MACRO));
    MacroList->Next = macro;
    macro = MacroList;
  }
  macro->Name = Xstrdup (name);
  macro->Content = Xstrdup (content);
  macro->ArgList = NULL;
  macro->LocalRecursion = 0;
}

/**
 * Creates a macro, that won't overwrite existing macros with the same name.
 * For the reason that it should only be used for macro paramters, no
 * argument list is necessary.
 */
void LocMacro (char *name, char *content, int current_recursion)
{
  MACRO *macro;

  macro = MacroList;
  MacroList = Xmalloc (sizeof (MACRO));
  MacroList->Next = macro;
  MacroList->Name = Xstrdup (name);
  MacroList->Content = Xstrdup (content);
  MacroList->ArgList = NULL;
  MacroList->LocalRecursion = current_recursion+1;
}

/**
 * Removes the macro that was lastly defined
 */
void RemoveLastMacro ()
{
  MACRO *macro;

  macro = MacroList->Next;
  FreeMacro (MacroList);
  MacroList = macro;
}

/**
 * Removes an equate/macro from the macro list
 */
void RemoveMacro (char *name)
{
  MACRO *last,
        *macro;

  if ((macro = getMacro (name)) == NULL)
    Error ("macro `%s' not exists");
  if (macro->LocalRecursion)
    Error ("`#undef' used with macro argument `%s'");

 /** macro is at the begin of the macro list */
  if (macro == MacroList)
  {
    RemoveLastMacro ();
    return;
  }

 /** search macro, that comes before the macro that is to be removed */
  for (last = MacroList; last->Next != macro; last = last->Next);

 /** remove the macro from the list */
  last->Next = macro->Next;
  FreeMacro (macro);
}

/**
 * Adds an argument to an argument-list. Automatically checks whether the
 * argument is an ellipse (arg...).
 */
void NewArgument (ARGLIST **arglist, char *name)
{
  int  ellipse = 0;

  while (*arglist != NULL)
  {
    if ((*arglist)->Ellipse)
      Error ("macro argument defined after ellipse");
    arglist = &((*arglist)->Next);
  }

  *arglist = Xmalloc (sizeof (ARGLIST));
  (*arglist)->Next = NULL;
  if (strlen (name) > 3 && !strcmp (strchr (name, '\0')-3, "..."))
  {
    ellipse = 1;
    (*arglist)->Name = Xstrdup (name);
    (*arglist)->Name[strlen (name)-3] = '\0';
  }
  else
    (*arglist)->Name = Xstrdup (name);
  (*arglist)->Ellipse = ellipse;
}

/***************************************************************************
 * Local symbol routines
 */

/**
 * Creates a new level for local symbols on the `LOCALSTACK *Local' stack
 */
void NewLocals ()
{
  LOCALSTACK *last,
             *locals;

  locals = Xmalloc (sizeof (LOCALSTACK));
  locals->Prepend = LocalPrependNum++;
  locals->LocalNum = 0;

  last = Locals;
  Locals = locals;
  Locals->Last = last;
}

/**
 * Adds a new local symbol to `LOCAL *LocalList'
 */
void NewLocalSymb (char *symb)
{
  LOCAL *local,
        *last;

  Locals->LocalNum++;

  local = Xmalloc (sizeof (LOCAL));
  local->OrigName = Xstrdup (symb);
  local->LocalName = Xmalloc (strlen (symb) + 8);
  sprintf (local->LocalName, "%s___%04i", symb, Locals->Prepend);

  last = LocalList;
  LocalList = local;
  LocalList->Next = last;
}

/**
 * Returns a pointer to the local name of a symbol, or NULL, if the symbol
 * isn't in the local symbols list
 */
char *getLocalSymb (char *symb)
{
  LOCAL *Iloc;

  for (Iloc = LocalList; Iloc != NULL; Iloc = Iloc->Next)
    if (!strcmp (symb, Iloc->OrigName))
      return (Iloc->LocalName);

  return (NULL);
}

/**
 * Removes the lastly defined local symbol
 */
void RemoveLastLocal ()
{
  LOCAL *next;

  free (LocalList->OrigName);
  free (LocalList->LocalName);
  next = LocalList->Next;
  free (LocalList);
  LocalList = next;
}

/**
 * Removes all locals from the current level
 */
void RemoveLocals ()
{
  int        I;
  LOCALSTACK *last;

  for (I = 0; I < Locals->LocalNum; I++)
    RemoveLastLocal ();

  last = Locals->Last;
  free (Locals);
  Locals = last;
}

/***************************************************************************
 * Input routines
 */

/**
 * Sets the current input (File/String). The older input will be continued,
 * when the current is done.
 */
void NewInput (FILE *hdl, char *str, char *file, int localmacros)
{
  INPUTSTACK *input,
             *last;

  if (Recursion++ >= MAXRECURSION)
    Error ("recursion overflow");

  input = Xmalloc (sizeof (INPUTSTACK));
  input->Hdl = hdl;
  input->File = Xstrdup (file);
  input->Str = str;
  input->LocalMacros = localmacros;
  input->LabelRead = 0;

  if (hdl)
    input->LineNum = 1;
  else
    input->LineNum = Input->LineNum;

  last = Input;
  Input = input;
  Input->Last = last;
}

/**
 * Restores the last input (is called, when an input is at the end)
 */
void LastInput ()
{
  INPUTSTACK *last;
  int        I;

  for (I = 0; I < Input->LocalMacros; I++)
    RemoveLastMacro ();

  last = Input->Last;
  free (Input->Str);
  free (Input);
  Input = last;

  Recursion--;
}

/**
 * Reads a char from the current input. EOF if no more input.
 */
int ReadChar ()
{
  int c;

  if (Input->Hdl != NULL)
  {
    if ((c = fgetc (Input->Hdl)) == '\n')
      Input->LineNum++;
    if (c == EOF)
      fclose (Input->Hdl);
  }
  else
    if (!(c = *Input->Str++))
      c = EOF;

  if (c == EOF)
  {
    if (Input->Last)
    {
      LastInput ();
      return (ReadChar ());
    }
    InputEnd = 1;
    return (EOF);
  }

  return (c);
}

/**
 * Pushs a char back to where it came from
 */
void UnRead (int c)
{
  if (Input->Hdl != NULL)
  {
    ungetc (c, Input->Hdl);
    if (c == '\n')
      Input->LineNum--;
  }
  else if (c != EOF)
    *--Input->Str = c;
}


/**
 * Returns, whether the given character can be a part of a multichar token.
 */
int istokchr (int c)
{
  return (isalnum (c) || strchr ("\\_.:@", c) != NULL);
}


char *__UnReadToken__ = NULL;
/**
 * Pushs a token back to where it came from
 */
void UnReadToken (char *token)
{
  __UnReadToken__ = Xstrdup (token);
}

/**
 * Reads a token from the current input
 */
char *ReadToken (char *token)
{
  char *ptr = token;
  int  c;

 /** restore a pushed token ? */
  if (__UnReadToken__)
  {
    strcpy (token, __UnReadToken__);
    free (__UnReadToken__);
    __UnReadToken__ = NULL;
    return (token);
  }

  if (Input->LabelRead)
  {
    *ptr++ = ':';
    Input->LabelRead = 0;
  }
  else if (istokchr (c = ReadChar ()))
  {
    *ptr++ = c;
    while (istokchr (c = ReadChar ()))
      *ptr++ = c;
    UnRead (c);
   /** label read -> return the colon the next time, ReadToken is called */
    if (*(ptr-1) == ':')
    {
      Input->LabelRead = 1;
      ptr--;
    }
  }
  else if (c == ',')
  {
    while ((c = ReadChar ()) == ' ' || c == '\t');
    UnRead (c);
    *ptr++ = ',';
  }
  else if (c == '"')
  {
    *ptr++ = c;
    while ((c = ReadChar ()) != '"' && c != EOF)
      *ptr++ = c;
    *ptr++ = c;
  }
  else if (c == '<')
  {
    *ptr++ = c;
    while ((c = ReadChar ()) != '>' && c != EOF)
      *ptr++ = c;
    *ptr++ = c;
  }
  else if (c == '\'')
  {
    *ptr++ = c;
    while ((c = ReadChar ()) != '\'' && c != EOF)
      *ptr++ = c;
    *ptr++ = c;
  }
  else if (c == '{')
  {
    int level = 1;

    while (level)
    {
      if ((c = ReadChar ()) == '{')
      {
        level++;
        *ptr++ = '{';
      }
      else if (c == '}')
      {
        if (--level)
          *ptr++ = '}';
      }
      else if (c == EOF)
        Error ("missing '}'");
      else
        *ptr++ = c;
    }
  }
  else if (c == ';')
  {
    while ((c = ReadChar ()) != '\n' && c != EOF);
    *ptr++ = '\n';
  }
  else if (c == '/')
  {
   /** C++ like comment: //..... */
    if ((c = ReadChar ()) == '/')
    {
      while ((c = ReadChar ()) != '\n' && c != EOF);
      *ptr++ = '\n';
    }
   /* C like comment: /+ ...+/ (where + should be *) */
    else if (c == '*')
    {
      int comment = 1;
      while ((c = ReadChar ()) != EOF && comment)
      {
        if (c == '*')
        {
          if ((c = ReadChar ()) == '/')
            comment--;
          else
            UnRead (c);
        }
        else if (c == '/')
          if ((c = ReadChar ()) == '*')
            comment++;
          else
            UnRead (c);
      }
      UnRead (c);
    }
    else
    {
      UnRead (c);
      *ptr++ = '/';
    }
  }
  else if (c == ' ' || c == '\t')
  {
    while ((c = ReadChar ()) == ' ' || c == '\t');
    UnRead (c);
    *ptr++ = '\t';
  }
  else if (c != '|')
    *ptr++ = c;

  *ptr = '\0';

  if (ptr = getLocalSymb (token))
  {
    return (strcpy (token, ptr));
  }
  return (token);
}


char *__UnReadTokenM__ = NULL;
/**
 * Pushs a token back to where it came from
 */
void UnReadTokenM (char *token)
{
  __UnReadTokenM__ = Xstrdup (token);
}

/**
 * Reads a token, and if a macro is found, sets the input to it.
 */
char *ReadTokenM (char *token)
{
  MACRO *macro;

 /** restore a pushed token ? */
  if (__UnReadTokenM__)
  {
    strcpy (token, __UnReadTokenM__);
    free (__UnReadTokenM__);
    __UnReadTokenM__ = NULL;
    return (token);
  }

 /** is the current token a macro ? */
  if (macro = getMacro (ReadToken (token)))
  {
   /** macro with parameters */
    if (macro->ArgList)
    {
      char    *content;
      int     parnum,              /* number of macro arguments */
              lparanthese,         /* number of '(' */
              crec = Recursion;    /* current recursion (in case it changes
                                      during argument passing due to macro
                                      expansions within the arguments) */
      ARGLIST *Iarg;
      content = Xmalloc (65536);

      if (*ReadTokenM (token) == '\t')
      {
        if (*ReadTokenM (token) != '(')
          Error ("`(' expected");
      }
      else if (*token != '(')
        Error ("`(' expected");

     /** read the macro parameters as new, local macros */
      for (Iarg = macro->ArgList, parnum = 0;
           Iarg != NULL; Iarg = Iarg->Next)
      {
        content[0] = '\0';
        for (lparanthese = 0;;)
        {
          ReadTokenM (token);
          if (!strcmp (token, "("))
            lparanthese++;
          else if (!strcmp (token, ")"))
            lparanthese--;
          if (lparanthese < 0
              || (!strcmp (token, ",") && !Iarg->Ellipse))
            break;
          strcat (content, token);
          if (InputEnd)
            Error ("unexpected end of file");
        }
        LocMacro (Iarg->Name, content, crec);
        parnum++;

        if (*token == ')')
          break;
      }
      if (*token == ')' && Iarg->Next && !Iarg->Next->Ellipse)
        Error ("too few arguments to macro `%s'", macro->Name);
      if (*token != ')')
        Error ("too many arguments to macro `%s'", macro->Name);

      NewInput (NULL, macro->Content, Input->File, parnum);
      free (content);
      return (ReadTokenM (token));
    }
   /** macro without parameters */
    else
    {
      NewInput (NULL, macro->Content, Input->File, 0);
      return (ReadTokenM (token));
    }
  }
  return (token);
}


/***************************************************************************
 * Routines to process preprocessor commands
 */

/**
 * Processes #include
 */
void Pre_include_ ()
{
  FILE *newhdl;
  char *nameend,
       drive[MAXDRIVE], path[MAXDIR], file[MAXFILE], ext[MAXEXT],
       fullname[MAXPATH];

  if (*ReadToken (Token) == '\t')
    ReadToken (Token);

  if ((Token[0] != '"' && Token[0] != '<') ||
      (Token[0] == '"' && (nameend = strchr (&Token[1], '"')) == NULL) ||
      (Token[0] == '<' && (nameend = strchr (&Token[1], '>')) == NULL))
    Error ("syntax");

  *nameend = '\0';

  fnsplit (&Token[1], NULL, NULL, file, ext);
  if (Token[0] == '<')
    fnsplit (IncludePath, drive, path, NULL, NULL);
  else
    fnsplit (SourcePath, drive, path, NULL, NULL);

  fnmerge (fullname, drive, path, file, ext);

  if ((newhdl = fopen (fullname, "rt")) == NULL)
    Error ("Can't open include-file `%s': %s\n\n", &Token[1], strerror (errno));

  fputc ('\n', OutHdl);
  NewInput (newhdl, NULL, fullname, 0);
}

/**
 * Processes #define
 */
void Pre_define_ ()
{
  char        name[256],
              *content;
  ARGLIST     *arglist = NULL;
  int         level,
              I,
              c;

  content = Xmalloc (65536);

 /** read macro name */
  if (*ReadToken (Token) != '\t')
    Error ("syntax");
  if (!istokchr (*ReadToken (Token)) || strlen (Token) >= 255)
    Error ("invalid macro name");
  strcpy (name, Token);

 /** read parameterlist */
  if ((c = ReadChar ()) == '(')
  {
    for (;;)
    {
      if (!istokchr (*ReadToken (Token)))
        Error ("invalid macro argument name: `%s'", Token);
      NewArgument (&arglist, Token);
      ReadToken (Token);
      if (Token[0] == ')')
        break;
      if (Token[0] == EOF)
        Error ("syntax");
      if (Token[0] != ',')
        Error ("`,' expected, `%s' found instead", Token);
    }
  }
  else
    UnRead (c);

 /** read macro content */
  /* skip spaces */
  while ((c = ReadChar ()) == ' ' || c == '\t');

  for (I = level = 0, content[0] = '\0'; c != '\n' || level; c = ReadChar ())
  {
    if (c == '{')
    {
      if (level++)
        content[I++] = '{';
    }
    else if (c == '}')
    {
      if (--level)
        content[I++] = '}';
    }
    else
      content[I++] = c;

    if (InputEnd)
      if (level)
        Error ("missing `}'");
      else
        Error ("unexpected end of file");
  }
  content[I] = '\0';
  UnRead ('\n');

  NewMacro (name, content, arglist);
  free (content);
}

/**
 * Processes #equ
 */
void Pre_equ_ ()
{
  char name[256],
       *content;

  content = Xmalloc (65536);

 /** read macro name */
  if (*ReadToken (Token) != '\t')
    Error ("syntax");
  if (!istokchr (*ReadToken (Token)) || strlen (Token) >= 255)
    Error ("invalid macro name");
  strcpy (name, Token);

 /** read macro content */
  /* skip spaces */
  while (!strcmp (ReadTokenM (Token), "\t"));

  for (content[0] = '\0'; strcmp (Token, "\n"); ReadTokenM (Token))
    strcat (content, Token);

  UnReadTokenM (Token);

  NewEquate (name, content);
  free (content);
}

/**
 * Processes #undef
 */
void Pre_undef_ ()
{
  char name[256];

 /** read macro name */
  if (*ReadToken (Token) != '\t')
    Error ("syntax");
  if (!istokchr (*ReadToken (Token)) || strlen (Token) >= 255)
    Error ("invalid macro name");
  strcpy (name, Token);

 /** remove macro */
  RemoveMacro (name);
}


enum IFTYPE {_IFDEF_, _IFNDEF_} iftype;
/**
 * Processes #ifdef/#ifndef
 */
void Pre_ifdef__ifndef_ (enum IFTYPE iftype)
{
 /** skip spaces */
  if (strcmp (ReadToken (Token), "\t"))
    Error ("syntax");
 /** read macro name */
  if (!istokchr (*ReadToken (Token)))
    Error ("invalid macro name: `%s'", Token);
 /** skip till #endif if macro defined/notdefined - depending on #if directive */
  if (iftype == _IFDEF_ ? (getMacro (Token) == NULL) :
                          (getMacro (Token) != NULL))
  {
    int level = 1,
        precommand = 0;       /* '#' read before ? */
    while (level)
    {
      ReadToken (Token);
      if (precommand)
        if (!strcmp (Token, "ifdef") || !strcmp (Token, "ifndef"))
          level++;
        else if (!strcmp (Token, "endif"))
          level--;
        else if (!strcmp (Token, "else") && level == 1)
        {
          IfLevel++;
          break;
        }
      if (!strcmp (Token, "#"))
        precommand = 1;
      else
        precommand = 0;
      if (InputEnd)
        Error ("missing `#endif'");
    }
  }
  else
    IfLevel++;
}

/**
 * Processes #else (skips everything till #endif)
 */
void Pre_else_ ()
{
  int level = 1,
      precommand = 0;       /* '#' read before ? */

  if (--IfLevel < 0)
    Error ("`#else' without `#ifdef' or `#ifndef'");

  while (level)
  {
    ReadToken (Token);
    if (precommand)
      if (!strcmp (Token, "ifdef") || !strcmp (Token, "ifndef"))
        level++;
      else if (!strcmp (Token, "endif"))
        level--;
    if (!strcmp (Token, "#"))
      precommand = 1;
    else
      precommand = 0;
    if (InputEnd)
      Error ("missing `#endif'");
  }
}

/**
 * Processes #endif
 */
void Pre_endif_ ()
{
  if (--IfLevel < 0)
    Error ("`#endif' without `#ifdef' or `#ifndef'");
}

/**
 * Processes #error
 */
void Pre_error_ ()
{
  char errorstring[1024];
  int I, c;

  for (I = 0; (c = ReadChar ()) != '\n'; I++)
  {
    errorstring[I] = c;
    if (I >= 1023)
      Error ("error message overflow");
  }
  errorstring[I] = '\0';
  UnRead ('\n');
  Error ("`#error%s'", errorstring);
  ReadChar ();
}

/**
 * Processes #local
 */
void Pre_local_ ()
{
  char name[256];
  int  I,
       c;

  NewLocals ();

 /** read the names of the local symbols */
  for (;;)
  {
   /** skip spaces */
    while ((c = ReadChar ()) == '\t' || c == ' ');

    for (I = 0; c != '\n' && c != ','; c = ReadChar (), I++)
      if (!istokchr (c) || c >= 255)
        Error ("invalid name for local symbol");
      else
        name[I] = c;

    name[I] = '\0';
    NewLocalSymb (name);

    if (c == '\n')
      break;
  }
  UnRead ('\n');
}

/**
 * Processes #endlocal
 */
void Pre_endlocal_ ()
{
  RemoveLocals ();
}

/**
 * Processes a precompiler command (e.g. #include, #define etc.)
 * return:
 *   0 = `command' is not a precompiler command!
 *   1 = precompiler command processed
 */
int ProcessPreCommand (char *command)
{
  if (!strcmp (command, "include"))
    Pre_include_ ();
  else if (!strcmp (command, "define"))
    Pre_define_ ();
  else if (!strcmp (command, "equ"))
    Pre_equ_ ();
  else if (!strcmp (command, "undef"))
    Pre_undef_ ();
  else if (!strcmp (command, "ifdef"))
    Pre_ifdef__ifndef_ (_IFDEF_);
  else if (!strcmp (command, "ifndef"))
    Pre_ifdef__ifndef_ (_IFNDEF_);
  else if (!strcmp (command, "else"))
    Pre_else_ ();
  else if (!strcmp (command, "endif"))
    Pre_endif_ ();
  else if (!strcmp (command, "error"))
    Pre_error_ ();
  else if (!strcmp (command, "local"))
    Pre_local_ ();
  else if (!strcmp (command, "endlocal"))
    Pre_endlocal_ ();
  else
    return (0);
  return (1);
}


/**
 * Precompiles a file until it ends, or a string (e.g. macro contents)
 */
void PreCompile ()
{
  int linebreak = 1,
      tabulator = 0;

  while (!InputEnd)
  {
    ReadTokenM (Token);
   /** linebreak - token */
    if (!strcmp (Token, "\n"))
    {
      if (!linebreak)
      {
        fprintf (OutHdl, "\t;%s(%i)\n", Input->File,
                                        Input->LineNum - (Input->Hdl ? 1 : 0));
        linebreak = 1;
      }
      tabulator = 0;
      continue;
    }
   /** preprocessor command? */
    if (!strcmp (Token, "#"))
      if (!ProcessPreCommand (ReadToken (Token)))
      {
        UnReadToken (Token);
        if (tabulator) fputs ("\t#", OutHdl), tabulator = 0;
        else fputc ('#', OutHdl);
        continue;
      }
      else
      {
        tabulator = 0;
        continue;
      }

   /** tabulator? */
    if (!strcmp (Token, "\t"))
      tabulator = 1;

   /** anything else */
    else if (Token[0])
    {
      if (tabulator) fputc ('\t', OutHdl);
      linebreak = 0;
      tabulator = 0;
      fputs (Token, OutHdl);
    }
  }
}


/**
 * Shows a short command line option summary
 */
void Help ()
{
  static char *helptext[] =
  {
    "",
    "Syntax: prefargo <sourcefile> [options]",
    "Options:",
    "  -o<filename> Specifies the output file. By default it will be gained from the",
    "               sourcefile's name, by changing the ending to `.asm' or, if the",
    "               sourcefile is allready an `.asm'-file, to `.as_'.",
    "  -i<filename> Specifies the path for #include <...> - commands",
    "  -s<filename> Specifies the path for #include \"...\" - commands",
    NULL
  };
  int I;

  for (I = 0; helptext[I]; I++)
    puts (helptext[I]);
}

/***************************************************************************
 * MAIN - processes command line arguments
 */
int main (int argn, char *argv[])
{
  FILE *inhdl;
  char *fargoenv;
  int  I;

  puts ("\n\t---------------------------------------------------------------");
  puts (  "\t                        Precompiler "VERSION_STR"                       ");
  puts (  "\t                Copyright (C) 1998 David Khling               ");
  puts (  "\t---------------------------------------------------------------\n");


  if (argn < 2)
  {
    fputs ("\nToo few arguments.\n\n", stderr);
    Help ();
    return (1);
  }

 /** process options */
  for (I = 2; I < argn; I++)
  {
    if (argv[I][0] != '-' && argv[I][0] != '/')
    {
      fprintf (stderr, "\nOption expected, `%s' found instead.\n\n", argv[I]);
      Help ();
      return (1);
    }
    switch (argv[I][1])
    {
      case 'i': IncludePath = &argv[I][2];
                break;
      case 's': SourcePath = &argv[I][2];
                break;
      case 'o': OutFile = &argv[I][2];
                break;
      default:  fprintf (stderr, "\nInvalid option: `%s'\n\n", argv[I]);
                Help ();
                return (1);
    }
  }

  if (fargoenv = getenv ("FARGO"))
  {
    if (!IncludePath)
    {
      IncludePath = Xmalloc (256);
      strcpy (IncludePath, fargoenv);
      strcat (IncludePath, "\\asm\\include\\");
    }
    if (!SourcePath)
    {
      SourcePath = Xmalloc (256);
      strcpy (SourcePath, fargoenv);
      strcat (SourcePath,  "\\asm\\prgm\\");
    }
  }

 /** if no outfile is specified -> generate one, by changing the extension */
  if (!OutFile)
  {
    char d[MAXDRIVE], p[MAXDIR], f[MAXFILE], e[MAXEXT];

    OutFile = Xmalloc (256);
    fnsplit (argv[1], d, p, f, e);
    if (!strcmp (e, "asm"))
      fnmerge (OutFile, d, p, f, "as_");
    else
      fnmerge (OutFile, d, p, f, "asm");
  }

  printf ("Precompiling `%s' to `%s'...\n", argv[1], OutFile);
  if (__file_exists (OutFile))
  {
    char str[256];
    printf ("Destination file allready exists. Overwrite (Y/N)? ");
    scanf ("%s", str);
    if (stricmp (str, "y"))
      return (1);
  }

 /** open/create the files */
  if ((inhdl = fopen (argv[1], "rt")) == NULL)
  {
    fprintf (stderr, "\nCan't open `%s': %s\n\n", argv[1], strerror (errno));
    return (1);
  }
  if ((OutHdl = fopen (OutFile, "wt")) == NULL)
  {
    fprintf (stderr, "\nCan't create `%s': %s\n\n", argv[2], strerror (errno));
    return (1);
  }

  NewInput (inhdl, NULL, argv[1], 0);
  PreCompile ();
  fclose (OutHdl);

  putchar ('\n');

  return (0);
}
